@@ -1,4 +1,4 @@
-# Copyright (C) 2019-2022  The Software Heritage developers
+# Copyright (C) 2019-2024  The Software Heritage developers
 # See the AUTHORS file at the top-level directory of this distribution
 # License: GNU General Public License version 3, or any later version
 # See top-level LICENSE file for more information
@@ -7,10 +7,12 @@ from collections import Counter
 import dataclasses
 import datetime
 import functools
+import importlib
 import itertools
 import logging
 import random
 from typing import (
+    TYPE_CHECKING,
     Any,
     Callable,
     Dict,
@@ -26,7 +28,8 @@ from typing import (
     cast,
 )
 
-from cassandra import ConsistencyLevel, CoordinationFailure
+from cassandra import ConsistencyLevel, CoordinationFailure, ReadTimeout, WriteTimeout
+from cassandra.auth import AuthProvider
 from cassandra.cluster import EXEC_PROFILE_DEFAULT, Cluster, ExecutionProfile, ResultSet
 from cassandra.concurrent import execute_concurrent_with_args
 from cassandra.policies import DCAwareRoundRobinPolicy, TokenAwarePolicy
@@ -39,6 +42,9 @@ from tenacity import (
     wait_random_exponential,
 )
 
+if TYPE_CHECKING:
+    from _typeshed import DataclassInstance
+
 from swh.core.utils import grouper
 from swh.model.model import (
     Content,
@@ -49,6 +55,7 @@ from swh.model.model import (
     TimestampWithTimezone,
 )
 from swh.model.swhids import CoreSWHID
+from swh.storage.exc import QueryTimeout
 from swh.storage.interface import ListOrder, TotalHashDict
 
 from ..utils import remove_keys
@@ -64,6 +71,7 @@ from .model import (
     MetadataAuthorityRow,
     MetadataFetcherRow,
     ObjectCountRow,
+    ObjectReferenceRow,
     OriginRow,
     OriginVisitRow,
     OriginVisitStatusRow,
@@ -95,6 +103,22 @@ BATCH_INSERT_MAX_SIZE = 1000
 logger = logging.getLogger(__name__)
 
 
+def _instantiate_auth_provider(configuration: Dict) -> AuthProvider:
+    local_config = dict(configuration)
+    cls = local_config.pop("cls", None)
+    if not cls:
+        raise ValueError(
+            "Configuration error: The cls property is mandatory "
+            " in the auth_provider configuration section"
+        )
+
+    (module_path, class_name) = cls.rsplit(".", 1)
+    module = importlib.import_module(module_path, package=__package__)
+    AuthProvider = getattr(module, class_name)
+
+    return AuthProvider(**local_config)
+
+
 def get_execution_profiles(
     consistency_level: str = "ONE",
 ) -> Dict[object, ExecutionProfile]:
@@ -119,9 +143,25 @@ def get_execution_profiles(
 
 
 def create_keyspace(
-    hosts: List[str], keyspace: str, port: int = 9042, *, durable_writes=True
+    hosts: List[str],
+    keyspace: str,
+    port: int = 9042,
+    *,
+    durable_writes=True,
+    auth_provider: Optional[Dict] = None,
 ):
-    cluster = Cluster(hosts, port=port, execution_profiles=get_execution_profiles())
+    auth_provider_inst: Optional[AuthProvider] = None
+    if auth_provider:
+        auth_provider_inst = _instantiate_auth_provider(auth_provider)
+
+    cluster = Cluster(
+        hosts,
+        port=port,
+        execution_profiles=get_execution_profiles(),
+        auth_provider=auth_provider_inst,
+        connect_timeout=30,
+        control_connection_timeout=30,
+    )
     session = cluster.connect()
     extra_params = ""
     if not durable_writes:
@@ -137,6 +177,7 @@ def create_keyspace(
     )
     session.execute('USE "%s"' % keyspace)
     for query in CREATE_TABLES_QUERIES:
+        logger.debug("Running:\n%s", query)
         session.execute(query)
 
 
@@ -266,6 +307,23 @@ def _prepared_select_token_range_statement(
     )
 
 
+def _prepared_delete_statement(
+    row_class: Type[BaseRow],
+) -> Callable[
+    [Callable[[TSelf, TArg, NamedArg(Any, "statement")], TRet]],  # noqa
+    Callable[[TSelf, TArg], TRet],
+]:
+    """Shorthand for using `_prepared_statement` for `DELETE FROM`
+    statements."""
+
+    where_conditions = " AND ".join(
+        f"{column} = ?" for column in row_class.PARTITION_KEY
+    )
+    return _prepared_statement(
+        f"DELETE FROM {{keyspace}}.{row_class.TABLE} WHERE {where_conditions}"
+    )
+
+
 def _next_bytes_value(value: bytes) -> bytes:
     """Returns the next bytes value by incrementing the integer
     representation of the provided value and converting it back
@@ -284,12 +342,24 @@ class CqlRunner:
     to Cassandra."""
 
     def __init__(
-        self, hosts: List[str], keyspace: str, port: int, consistency_level: str
+        self,
+        hosts: List[str],
+        keyspace: str,
+        port: int,
+        consistency_level: str,
+        auth_provider: Optional[Dict] = None,
     ):
+        auth_provider_impl: Optional[AuthProvider] = None
+        if auth_provider:
+            auth_provider_impl = _instantiate_auth_provider(auth_provider)
+
         self._cluster = Cluster(
             hosts,
             port=port,
+            auth_provider=auth_provider_impl,
             execution_profiles=get_execution_profiles(consistency_level),
+            connect_timeout=30,
+            control_connection_timeout=30,
         )
         self.keyspace = keyspace
         self._session = self._cluster.connect()
@@ -318,27 +388,43 @@ class CqlRunner:
         stop=stop_after_attempt(MAX_RETRIES),
         retry=retry_if_exception_type(CoordinationFailure),
     )
-    def _execute_with_retries(self, statement, args: Optional[Sequence]) -> ResultSet:
+    def _execute_with_retries_inner(
+        self, statement, args: Optional[Sequence]
+    ) -> ResultSet:
         return self._session.execute(statement, args, timeout=1000.0)
 
+    def _execute_with_retries(self, statement, args: Optional[Sequence]) -> ResultSet:
+        try:
+            return self._execute_with_retries_inner(statement, args)
+        except (ReadTimeout, WriteTimeout) as e:
+            raise QueryTimeout(*e.args) from None
+
     @retry(
         wait=wait_random_exponential(multiplier=1, max=10),
         stop=stop_after_attempt(MAX_RETRIES),
         retry=retry_if_exception_type(CoordinationFailure),
     )
-    def _execute_many_with_retries(
+    def _execute_many_with_retries_inner(
         self, statement, args_list: Sequence[Tuple]
     ) -> Iterable[Dict[str, Any]]:
         for res in execute_concurrent_with_args(self._session, statement, args_list):
             yield from res.result_or_exc
 
-    def _add_one(self, statement, obj: BaseRow) -> None:
+    def _execute_many_with_retries(
+        self, statement, args_list: Sequence[Tuple]
+    ) -> Iterable[Dict[str, Any]]:
+        try:
+            return self._execute_many_with_retries_inner(statement, args_list)
+        except (ReadTimeout, WriteTimeout) as e:
+            raise QueryTimeout(*e.args) from None
+
+    def _add_one(self, statement, obj: "DataclassInstance") -> None:
         self._execute_with_retries(statement, dataclasses.astuple(obj))
 
     def _add_many(self, statement, objs: Sequence[BaseRow]) -> None:
         tables = {obj.TABLE for obj in objs}
         assert len(tables) == 1, f"Cannot insert to multiple tables: {tables}"
-        rows = list(map(dataclasses.astuple, objs))
+        rows = [dataclasses.astuple(cast("DataclassInstance", obj)) for obj in objs]
         for _ in self._execute_many_with_retries(statement, rows):
             # Need to consume the generator to actually run the INSERTs
             pass
@@ -356,7 +442,7 @@ class CqlRunner:
             # the row with the smallest token
             rows = self._execute_with_retries(statement, [TOKEN_BEGIN])
         if rows:
-            return row_class.from_dict(rows.one())  # type: ignore
+            return row_class.from_dict(rows.one())
         else:
             return None
 
@@ -451,7 +537,7 @@ class CqlRunner:
         self, ids: List[bytes], *, statement
     ) -> Iterator[Tuple[bytes, bytes, bytes, bytes]]:
         for row in self._execute_with_retries(statement, [ids]):
-            yield tuple(row[algo] for algo in HASH_ALGORITHMS)  # type: ignore
+            yield tuple(row[algo] for algo in HASH_ALGORITHMS)
 
     @_prepared_select_statement(
         ContentRow, f"WHERE token({', '.join(ContentRow.PARTITION_KEY)}) = ?"
@@ -478,6 +564,19 @@ class CqlRunner:
             for row in self._execute_with_retries(statement, [start, end, limit])
         )
 
+    @_prepared_delete_statement(ContentRow)
+    def content_delete(self, content_hashes: TotalHashDict, *, statement) -> None:
+        self._execute_with_retries(
+            statement,
+            [cast(dict, content_hashes)[algo] for algo in ContentRow.PARTITION_KEY],
+        )
+        for algo in HASH_ALGORITHMS:
+            table = content_index_table_name(algo, skipped_content=False)
+            query = f"""
+                DELETE FROM {self.keyspace}.{table} WHERE {algo} = %s
+            """
+            self._execute_with_retries(query, [cast(dict, content_hashes)[algo]])
+
     ##########################
     # 'content_by_*' tables
     ##########################
@@ -498,7 +597,7 @@ class CqlRunner:
         table = content_index_table_name(algo, skipped_content=False)
         query = f"SELECT target_token FROM {self.keyspace}.{table} WHERE {algo} = %s"
         return (
-            row["target_token"]  # type: ignore
+            row["target_token"]
             for row in self._execute_many_with_retries(
                 query, [(hash_,) for hash_ in hashes]
             )
@@ -577,6 +676,23 @@ class CqlRunner:
             SkippedContentRow.from_dict, self._execute_with_retries(statement, [token])
         )
 
+    @_prepared_delete_statement(SkippedContentRow)
+    def skipped_content_delete(
+        self, skipped_content_hashes: TotalHashDict, *, statement
+    ) -> None:
+        self._execute_with_retries(
+            statement,
+            [cast(dict, skipped_content_hashes)[algo] for algo in HASH_ALGORITHMS],
+        )
+        for algo in HASH_ALGORITHMS:
+            table = content_index_table_name(algo, skipped_content=True)
+            query = f"""
+                DELETE FROM {self.keyspace}.{table} WHERE {algo} = %s
+            """
+            self._execute_with_retries(
+                query, [cast(dict, skipped_content_hashes)[algo]]
+            )
+
     ##########################
     # 'skipped_content_by_*' tables
     ##########################
@@ -643,6 +759,10 @@ class CqlRunner:
             for row in self._execute_with_retries(statement, [start, end, limit])
         )
 
+    @_prepared_delete_statement(DirectoryRow)
+    def directory_delete(self, directory_id: bytes, *, statement) -> None:
+        self._execute_with_retries(statement, [directory_id])
+
     ##########################
     # 'directory_entry' table
     ##########################
@@ -714,6 +834,10 @@ class CqlRunner:
             self._execute_with_retries(statement, [directory_id, from_, limit]),
         )
 
+    @_prepared_delete_statement(DirectoryEntryRow)
+    def directory_entry_delete(self, directory_id: Sha1Git, *, statement) -> None:
+        self._execute_with_retries(statement, [directory_id])
+
     ##########################
     # 'revision' table
     ##########################
@@ -754,6 +878,10 @@ class CqlRunner:
             for row in self._execute_with_retries(statement, [start, end, limit])
         )
 
+    @_prepared_delete_statement(RevisionRow)
+    def revision_delete(self, revision_id: Sha1Git, *, statement) -> None:
+        self._execute_with_retries(statement, [revision_id])
+
     ##########################
     # 'revision_parent' table
     ##########################
@@ -773,6 +901,10 @@ class CqlRunner:
             for row in self._execute_with_retries(statement, [revision_id])
         )
 
+    @_prepared_delete_statement(RevisionParentRow)
+    def revision_parent_delete(self, revision_id: Sha1Git, *, statement) -> None:
+        self._execute_with_retries(statement, [revision_id])
+
     ##########################
     # 'release' table
     ##########################
@@ -807,6 +939,10 @@ class CqlRunner:
             for row in self._execute_with_retries(statement, [start, end, limit])
         )
 
+    @_prepared_delete_statement(ReleaseRow)
+    def release_delete(self, release_id: Sha1Git, *, statement) -> None:
+        self._execute_with_retries(statement, [release_id])
+
     ##########################
     # 'snapshot' table
     ##########################
@@ -833,6 +969,10 @@ class CqlRunner:
             for row in self._execute_with_retries(statement, [start, end, limit])
         )
 
+    @_prepared_delete_statement(SnapshotRow)
+    def snapshot_delete(self, snapshot_id: Sha1Git, *, statement) -> None:
+        self._execute_with_retries(statement, [snapshot_id])
+
     ##########################
     # 'snapshot_branch' table
     ##########################
@@ -935,7 +1075,7 @@ class CqlRunner:
         branch_name_exclude_prefix: Optional[bytes] = None,
     ) -> Iterable[SnapshotBranchRow]:
         prefix = branch_name_exclude_prefix
-        if prefix is None:
+        if prefix is None or (from_ > prefix and not from_.startswith(prefix)):
             return self.snapshot_branch_get_from_name(snapshot_id, from_, limit)
         else:
             # get branches before the exclude prefix
@@ -954,6 +1094,10 @@ class CqlRunner:
                 )
             return branches
 
+    @_prepared_delete_statement(SnapshotBranchRow)
+    def snapshot_branch_delete(self, snapshot_id: Sha1Git, *, statement) -> None:
+        self._execute_with_retries(statement, [snapshot_id])
+
     ##########################
     # 'origin' table
     ##########################
@@ -1032,6 +1176,10 @@ class CqlRunner:
 
         return next_id
 
+    @_prepared_delete_statement(OriginRow)
+    def origin_delete(self, sha1: bytes, *, statement) -> None:
+        self._execute_with_retries(statement, [sha1])
+
     ##########################
     # 'origin_visit' table
     ##########################
@@ -1118,6 +1266,10 @@ class CqlRunner:
         yield from self._origin_visit_iter_from(start_token)
         yield from self._origin_visit_iter_to(start_token)
 
+    @_prepared_delete_statement(OriginVisitRow)
+    def origin_visit_delete(self, origin_url: str, *, statement) -> None:
+        self._execute_with_retries(statement, [origin_url])
+
     ##########################
     # 'origin_visit_status' table
     ##########################
@@ -1176,7 +1328,6 @@ class CqlRunner:
         *,
         statement,
     ) -> Iterable[OriginVisitStatusRow]:
-
         args = (origin_url, visit_from, visit_to)
 
         return map(
@@ -1223,6 +1374,10 @@ class CqlRunner:
             if d["snapshot"] is not None
         }
 
+    @_prepared_delete_statement(OriginVisitStatusRow)
+    def origin_visit_status_delete(self, origin_url: str, *, statement) -> None:
+        self._execute_with_retries(statement, [origin_url])
+
     #########################
     # 'raw_extrinsic_metadata_by_id' table
     #########################
@@ -1524,6 +1679,30 @@ class CqlRunner:
             for row in self._execute_with_retries(statement, [target_type, target])
         )
 
+    ##########################
+    # 'object_references' table
+    ##########################
+
+    @_prepared_insert_statement(ObjectReferenceRow)
+    def object_reference_add_concurrent(
+        self, entries: List[ObjectReferenceRow], *, statement
+    ) -> None:
+        if len(entries) == 0:
+            # nothing to do
+            return
+        self._add_many(statement, entries)
+
+    @_prepared_select_statement(
+        ObjectReferenceRow, "WHERE target_type = ? AND target = ? LIMIT ?"
+    )
+    def object_reference_get(
+        self, target: Sha1Git, target_type: str, limit: int, *, statement
+    ) -> Iterable[ObjectReferenceRow]:
+        return map(
+            ObjectReferenceRow.from_dict,
+            self._execute_with_retries(statement, [target_type, target, limit]),
+        )
+
     ##########################
     # Miscellaneous
     ##########################
