@@ -1,4 +1,4 @@
-# Copyright (C) 2019-2021  The Software Heritage developers
+# Copyright (C) 2019-2024  The Software Heritage developers
 # See the AUTHORS file at the top-level directory of this distribution
 # License: GNU General Public License version 3, or any later version
 # See top-level LICENSE file for more information
@@ -77,7 +77,6 @@ CREATE TABLE IF NOT EXISTS content (
     blake2s256    blob,
     length        bigint,
     ctime         timestamp,
-        -- creation time, i.e. time of (first) injection into the storage
     status        ascii,
     PRIMARY KEY ((sha256), sha1, sha1_git, blake2s256)
 );""",
@@ -89,7 +88,6 @@ CREATE TABLE IF NOT EXISTS skipped_content (
     blake2s256    blob,
     length        bigint,
     ctime         timestamp,
-        -- creation time, i.e. time of (first) injection into the storage
     status        ascii,
     reason        text,
     origin        text,
@@ -97,35 +95,30 @@ CREATE TABLE IF NOT EXISTS skipped_content (
 );""",
     """
 CREATE TABLE IF NOT EXISTS revision (
-    id                              blob PRIMARY KEY,
+    id                              blob,
     date                            microtimestamp_with_timezone,
     committer_date                  microtimestamp_with_timezone,
     type                            ascii,
-    directory                       blob,  -- source code "root" directory
+    directory                       blob,
     message                         blob,
     author                          person,
     committer                       person,
     synthetic                       boolean,
-        -- true iff revision has been created by Software Heritage
     metadata                        text,
-        -- extra metadata as JSON(tarball checksums, etc...)
     extra_headers                   frozen<list <list<blob>> >,
-        -- extra commit information as (tuple(key, value), ...)
     raw_manifest                    blob,
-        -- NULL if the object can be rebuild from other cells and revision_parent.
+    PRIMARY KEY ((id))
 );""",
     """
 CREATE TABLE IF NOT EXISTS revision_parent (
     id                     blob,
     parent_rank                     int,
-        -- parent position in merge commits, 0-based
     parent_id                       blob,
     PRIMARY KEY ((id), parent_rank)
 );""",
     """
-CREATE TABLE IF NOT EXISTS release
-(
-    id                              blob PRIMARY KEY,
+CREATE TABLE IF NOT EXISTS release (
+    id                              blob,
     target_type                     ascii,
     target                          blob,
     date                            microtimestamp_with_timezone,
@@ -133,32 +126,30 @@ CREATE TABLE IF NOT EXISTS release
     message                         blob,
     author                          person,
     synthetic                       boolean,
-        -- true iff release has been created by Software Heritage
     raw_manifest                    blob,
-        -- NULL if the object can be rebuild from other cells
+    PRIMARY KEY ((id))
 );""",
     """
 CREATE TABLE IF NOT EXISTS directory (
-    id              blob PRIMARY KEY,
-    raw_manifest                    blob
-        -- NULL if the object can be rebuild from (sorted) entries
+    id              blob,
+    raw_manifest                    blob,
+    PRIMARY KEY ((id))
 );""",
     """
 CREATE TABLE IF NOT EXISTS directory_entry (
     directory_id    blob,
-    name            blob,  -- path name, relative to containing dir
+    name            blob,
     target          blob,
-    perms           int,   -- unix-like permissions
-    type            ascii, -- target type
+    perms           int,
+    type            ascii,
     PRIMARY KEY ((directory_id), name)
 );""",
     """
 CREATE TABLE IF NOT EXISTS snapshot (
-    id              blob PRIMARY KEY,
+    id              blob,
+    PRIMARY KEY ((id))
 );""",
     """
--- For a given snapshot_id, branches are sorted by their name,
--- allowing easy pagination.
 CREATE TABLE IF NOT EXISTS snapshot_branch (
     snapshot_id     blob,
     name            blob,
@@ -189,12 +180,10 @@ WITH CLUSTERING ORDER BY (visit DESC, date DESC)
 ;""",  # 'WITH CLUSTERING ORDER BY' is optional with Cassandra 4, but ScyllaDB needs it
     """
 CREATE TABLE IF NOT EXISTS origin (
-    sha1            blob PRIMARY KEY,
+    sha1            blob,
     url             text,
     next_visit_id   int,
-        -- We need integer visit ids for compatibility with the pgsql
-        -- storage, so we're using lightweight transactions with this trick:
-        -- https://stackoverflow.com/a/29391877/539465
+    PRIMARY KEY ((sha1))
 );""",
     """
 CREATE TABLE IF NOT EXISTS metadata_authority (
@@ -208,25 +197,23 @@ CREATE TABLE IF NOT EXISTS metadata_fetcher (
     version         ascii,
     PRIMARY KEY ((name), version)
 );""",
+    # Cassandra handles null values for row properties as removals (tombstones), which
+    # are never cleaned up as the values were never set. To avoid this issue, we instead
+    # store a known invalid value of the proper type as a placeholder for null
+    # properties: for strings and bytes: the empty value; for visit ids (integers): 0.
+    # See comments for visit, snapshot, release, revision, path and directory fields
     """
 CREATE TABLE IF NOT EXISTS raw_extrinsic_metadata (
     id              blob,
-
     type            text,
     target          text,
-
-    -- metadata source
     authority_type  text,
     authority_url   text,
     discovery_date  timestamp,
     fetcher_name    ascii,
     fetcher_version ascii,
-
-    -- metadata itself
     format          ascii,
     metadata        blob,
-
-    -- context
     origin          text,
     visit           bigint,
     snapshot        text,
@@ -234,34 +221,7 @@ CREATE TABLE IF NOT EXISTS raw_extrinsic_metadata (
     revision        text,
     path            blob,
     directory       text,
-
     PRIMARY KEY ((target), authority_type, authority_url, discovery_date, id)
-
-    -- An explanation is in order for this primary key:
-    --
-    -- Intuitively, the primary key should only be 'id', because two metadata
-    -- entries are the same iff the id is the same; and 'id' is used for
-    -- deduplication.
-    --
-    -- However, we also want to query by
-    -- (target, authority_type, authority_url, discovery_date)
-    -- The naive solution to this would be an extra table, to use as index;
-    -- but it means 1. extra code to keep them in sync 2. overhead when writing
-    -- 3. overhead + random reads (instead of linear) when reading.
-    --
-    -- Therefore, we use a single table for both, by adding the column
-    -- we want to query with before the id.
-    -- It solves both a) the query/order issues and b) the uniqueness issue because:
-    --
-    -- a) adding the id at the end of the primary key does not change the rows' order:
-    --    for two different rows, id1 != id2, so
-    --    (target1, ..., date1) < (target2, ..., date2)
-    --    <=> (target1, ..., date1, id1) < (target2, ..., date2, id2)
-    --
-    -- b) the id is a hash of all the columns, so:
-    --    rows are the same
-    --    <=> id1 == id2
-    --    <=> (target1, ..., date1, id1) == (target2, ..., date2, id2)
 );""",
     """
 CREATE TABLE IF NOT EXISTS raw_extrinsic_metadata_by_id (
@@ -284,8 +244,16 @@ CREATE TABLE IF NOT EXISTS extid (
 CREATE TABLE IF NOT EXISTS extid_by_target (
     target_type     ascii,
     target          blob,
-    target_token    bigint, -- value of token(pk) on the "primary" table
+    target_token    bigint,
     PRIMARY KEY ((target_type, target), target_token)
+);""",
+    """
+CREATE TABLE IF NOT EXISTS object_references (
+    target_type     ascii,
+    target          blob,
+    source_type     ascii,
+    source          blob,
+    PRIMARY KEY ((target_type, target), source_type, source)
 );""",
 ]
 
@@ -297,6 +265,7 @@ CREATE TABLE IF NOT EXISTS content_by_{main_algo} (
     PRIMARY KEY (({main_algo}), target_token)
 );
 
+-- Secondary table, used for looking up "skipped_content" from a single hash
 CREATE TABLE IF NOT EXISTS skipped_content_by_{main_algo} (
     {main_algo}   blob,
     target_token  bigint, -- value of token(pk) on the "primary" table
@@ -322,6 +291,7 @@ TABLES = [
     "metadata_fetcher",
     "extid",
     "extid_by_target",
+    "object_references",
 ]
 
 HASH_ALGORITHMS = ["sha1", "sha1_git", "sha256", "blake2s256"]
