@@ -1,4 +1,4 @@
-# Copyright (C) 2019-2022  The Software Heritage developers
+# Copyright (C) 2019-2024  The Software Heritage developers
 # See the AUTHORS file at the top-level directory of this distribution
 # License: GNU General Public License version 3, or any later version
 # See top-level LICENSE file for more information
@@ -58,14 +58,16 @@ from swh.storage.interface import (
     VISIT_STATUSES,
     HashDict,
     ListOrder,
+    ObjectReference,
     OriginVisitWithStatuses,
     PagedResult,
     PartialBranches,
     Sha1,
+    SnapshotBranchByNameResponse,
     TotalHashDict,
 )
 from swh.storage.objstorage import ObjStorage
-from swh.storage.utils import map_optional, now
+from swh.storage.utils import now
 from swh.storage.writer import JournalWriter
 
 from . import converters
@@ -163,12 +165,13 @@ class CassandraStorage:
         self,
         hosts,
         keyspace,
-        objstorage,
+        objstorage=None,
         port=9042,
         journal_writer=None,
         allow_overwrite=False,
         consistency_level="ONE",
         directory_entries_insert_algo="one-by-one",
+        auth_provider: Optional[Dict] = None,
     ):
         """
         A backend of swh-storage backed by Cassandra
@@ -176,7 +179,8 @@ class CassandraStorage:
         Args:
             hosts: Seed Cassandra nodes, to start connecting to the cluster
             keyspace: Name of the Cassandra database to use
-            objstorage: Passed as argument to :class:`ObjStorage`
+            objstorage: Passed as argument to :class:`ObjStorage`; if unset,
+               use a NoopObjStorage
             port: Cassandra port
             journal_writer: Passed as argument to :class:`JournalWriter`
             allow_overwrite: Whether ``*_add`` functions will check if an object
@@ -193,11 +197,20 @@ class CassandraStorage:
                 * one-by-one: naive, one INSERT per directory entry, serialized
                 * concurrent: one INSERT per directory entry, concurrent
                 * batch: using UNLOGGED BATCH to insert many entries in a few statements
+            auth_provider: An optional dict describing the authentication provider to use.
+                Must contain at least a ``cls`` entry and the parameters to pass to the
+                constructor. For example::
+
+                    auth_provider:
+                        cls: cassandra.auth.PlainTextAuthProvider
+                        username: myusername
+                        password: mypassword
         """
         self._hosts = hosts
         self._keyspace = keyspace
         self._port = port
         self._consistency_level = consistency_level
+        self._auth_provider = auth_provider
         self._set_cql_runner()
         self.journal_writer: JournalWriter = JournalWriter(journal_writer)
         self.objstorage: ObjStorage = ObjStorage(self, objstorage)
@@ -225,7 +238,11 @@ class CassandraStorage:
     def _set_cql_runner(self):
         """Used by tests when they need to reset the CqlRunner"""
         self._cql_runner: CqlRunner = CqlRunner(
-            self._hosts, self._keyspace, self._port, self._consistency_level
+            self._hosts,
+            self._keyspace,
+            self._port,
+            self._consistency_level,
+            self._auth_provider,
         )
 
     def check_config(self, *, check_write: bool) -> bool:
@@ -312,7 +329,7 @@ class CassandraStorage:
                                 )
 
                     if collisions:
-                        collisions.append(content.hashes())
+                        collisions.append(dict(content.hashes()))
                         raise HashCollision(algo, content.get_hash(algo), collisions)
 
             (token, insertion_finalizer) = self._cql_runner.content_add_prepare(
@@ -561,6 +578,32 @@ class CassandraStorage:
         contents = [attr.evolve(c, ctime=now()) for c in content]
         return self._skipped_content_add(contents)
 
+    def skipped_content_find(self, content: HashDict) -> List[SkippedContent]:
+        if not set(content).intersection(HASH_ALGORITHMS):
+            raise StorageArgumentException(
+                "content keys must contain at least one "
+                f"of: {', '.join(sorted(HASH_ALGORITHMS))}"
+            )
+        # get first algo that was given
+        algo, hash_ = next(iter(cast(Dict[str, bytes], content).items()))
+        # because of collisions, we might get multiple tokens for different skipped contents
+        tokens = self._cql_runner.skipped_content_get_tokens_from_single_hash(
+            algo, hash_
+        )
+        # that means that now we need to filter out those that do not match the search criteria…
+        results = []
+        for token in tokens:
+            # and the token might actually also correspond to multiple,
+            # unrelated skipped content
+            for row in self._cql_runner.skipped_content_get_from_token(token):
+                row_d = row.to_dict()
+                if all(row_d[algo] == hash_ for algo, hash_ in content.items()):
+                    row_d["ctime"] = row_d["ctime"].replace(
+                        tzinfo=datetime.timezone.utc
+                    )
+                    results.append(SkippedContent(**row_d))
+        return results
+
     def skipped_content_missing(
         self, contents: List[Dict[str, Any]]
     ) -> Iterable[Dict[str, Any]]:
@@ -787,7 +830,7 @@ class CassandraStorage:
             revobject = converters.revision_to_db(revision)
             if revobject:
                 # Add parents first
-                for (rank, parent) in enumerate(revision.parents):
+                for rank, parent in enumerate(revision.parents):
                     self._cql_runner.revision_parent_add_one(
                         RevisionParentRow(
                             id=revobject.id, parent_rank=rank, parent_id=parent
@@ -987,7 +1030,7 @@ class CassandraStorage:
             self.journal_writer.snapshot_add([snapshot])
 
             # Add branches
-            for (branch_name, branch) in snapshot.branches.items():
+            for branch_name, branch in snapshot.branches.items():
                 if branch is None:
                     target_type: Optional[str] = None
                     target: Optional[bytes] = None
@@ -1086,7 +1129,10 @@ class CassandraStorage:
 
             branches_from = new_branches[-1].name
 
-            new_branches_filtered = new_branches
+            if len(new_branches) > branches_count:
+                new_branches_filtered = new_branches[:-1]
+            else:
+                new_branches_filtered = new_branches
 
             # Filter by target_type
             if target_types:
@@ -1114,6 +1160,7 @@ class CassandraStorage:
                 break
 
         if len(branches) > branches_count:
+            branches = branches[: branches_count + 1]
             last_branch = branches.pop(-1).name
         else:
             last_branch = None
@@ -1121,10 +1168,12 @@ class CassandraStorage:
         return PartialBranches(
             id=snapshot_id,
             branches={
-                branch.name: None
-                if branch.target is None
-                else SnapshotBranch(
-                    target=branch.target, target_type=TargetType(branch.target_type)
+                branch.name: (
+                    None
+                    if branch.target is None
+                    else SnapshotBranch(
+                        target=branch.target, target_type=TargetType(branch.target_type)
+                    )
                 )
                 for branch in branches
             },
@@ -1136,6 +1185,59 @@ class CassandraStorage:
         assert snapshot, "Could not find any snapshot"
         return snapshot.id
 
+    def snapshot_branch_get_by_name(
+        self,
+        snapshot_id: Sha1Git,
+        branch_name: bytes,
+        follow_alias_chain: bool = True,
+        max_alias_chain_length: int = 100,
+    ) -> Optional[SnapshotBranchByNameResponse]:
+        if self._cql_runner.snapshot_missing([snapshot_id]):
+            return None
+
+        resolve_chain: List[bytes] = []
+        while True:
+            branches = list(
+                self._cql_runner.snapshot_branch_get_from_name(
+                    snapshot_id=snapshot_id, from_=branch_name, limit=1
+                )
+            )
+            if len(branches) != 1 or branches[0].name != branch_name:
+                # target branch is None, there could be items in aliases_followed
+                target = None
+                break
+            branch = branches[0]
+            resolve_chain.append(branch_name)
+            if branch.target_type != TargetType.ALIAS.value or not follow_alias_chain:
+                # first non alias branch or the first branch when follow_alias_chain is False
+                target = (
+                    SnapshotBranch(
+                        target=branch.target,
+                        target_type=TargetType(branch.target_type),
+                    )
+                    if branch.target
+                    else None
+                )
+                break
+            elif (
+                # Circular reference
+                resolve_chain.count(branch_name) > 1
+                # Too many re-directs
+                or len(resolve_chain) >= max_alias_chain_length
+            ):
+                target = None
+                break
+            # Branch has a non-None target with type alias
+            assert branch.target is not None
+            branch_name = branch.target
+
+        return SnapshotBranchByNameResponse(
+            # resolve_chian has items, brach_found must be True
+            branch_found=bool(resolve_chain),
+            target=target,
+            aliases_followed=resolve_chain,
+        )
+
     ##########################
     # Origin
     ##########################
@@ -1173,7 +1275,7 @@ class CassandraStorage:
 
         origins = []
         # Take one more origin so we can reuse it as the next page token if any
-        for (tok, row) in self._cql_runner.origin_list(start_token, limit + 1):
+        for tok, row in self._cql_runner.origin_list(start_token, limit + 1):
             origins.append(Origin(url=row.url))
             # keep reference of the last id for pagination purposes
             last_id = tok
@@ -1478,12 +1580,16 @@ class CassandraStorage:
         return PagedResult(results=visit_statuses, next_page_token=next_page_token)
 
     def origin_visit_find_by_date(
-        self, origin: str, visit_date: datetime.datetime
+        self, origin: str, visit_date: datetime.datetime, type: Optional[str] = None
     ) -> Optional[OriginVisit]:
         # Iterator over all the visits of the origin
         # This should be ok for now, as there aren't too many visits
         # per origin.
-        rows = list(self._cql_runner.origin_visit_iter_all(origin))
+        rows = [
+            visit
+            for visit in self._cql_runner.origin_visit_iter_all(origin)
+            if type is None or visit.type == type
+        ]
 
         def key(visit):
             dt = visit.date.replace(tzinfo=datetime.timezone.utc) - visit_date
@@ -1588,7 +1694,7 @@ class CassandraStorage:
             ("directory", self._cql_runner.directory_missing),
         ]
 
-        for (object_type, query_fn) in queries:
+        for object_type, query_fn in queries:
             found_ids = missing_ids - set(query_fn(list(missing_ids)))
             for sha1_git in found_ids:
                 results[sha1_git].append(
@@ -1631,6 +1737,7 @@ class CassandraStorage:
     def raw_extrinsic_metadata_add(
         self, metadata: List[RawExtrinsicMetadata]
     ) -> Dict[str, int]:
+        """Add extrinsic metadata on objects (contents, directories, ...)."""
         self.journal_writer.raw_extrinsic_metadata_add(metadata)
         counter = Counter[ExtendedObjectType]()
         for metadata_entry in metadata:
@@ -1655,14 +1762,30 @@ class CassandraStorage:
                     fetcher_version=metadata_entry.fetcher.version,
                     format=metadata_entry.format,
                     metadata=metadata_entry.metadata,
-                    origin=metadata_entry.origin,
-                    visit=metadata_entry.visit,
-                    snapshot=map_optional(str, metadata_entry.snapshot),
-                    release=map_optional(str, metadata_entry.release),
-                    revision=map_optional(str, metadata_entry.revision),
-                    path=metadata_entry.path,
-                    directory=map_optional(str, metadata_entry.directory),
+                    # Cassandra handles null values for row properties as removals
+                    # (tombstones), which are never cleaned up as the values were never
+                    # set. To avoid this issue, we instead store a known invalid value
+                    # of the proper type as a placeholder for null properties: for
+                    # strings and bytes: the empty value; for visit ids (integers): 0.
+                    # The inverse conversion is performed when reading the data back
+                    # from Cassandra.
+                    origin=metadata_entry.origin or "",
+                    visit=metadata_entry.visit or 0,
+                    snapshot=str(metadata_entry.snapshot)
+                    if metadata_entry.snapshot
+                    else "",
+                    release=str(metadata_entry.release)
+                    if metadata_entry.release
+                    else "",
+                    revision=str(metadata_entry.revision)
+                    if metadata_entry.revision
+                    else "",
+                    path=metadata_entry.path or b"",
+                    directory=str(metadata_entry.directory)
+                    if metadata_entry.directory
+                    else "",
                 )
+
             except TypeError as e:
                 raise StorageArgumentException(*e.args)
 
@@ -1933,6 +2056,176 @@ class CassandraStorage:
             )
         return result
 
+    #########################
+    # 'object_references' table
+    #########################
+
+    def object_find_recent_references(
+        self, target_swhid: ExtendedSWHID, limit: int
+    ) -> List[ExtendedSWHID]:
+        rows = self._cql_runner.object_reference_get(
+            target=target_swhid.object_id,
+            target_type=target_swhid.object_type.value,
+            limit=limit,
+        )
+        return [converters.row_to_object_reference(row).source for row in rows]
+
+    def object_references_add(
+        self, references: List[ObjectReference]
+    ) -> Dict[str, int]:
+        to_add = list(
+            {converters.object_reference_to_row(reference) for reference in references}
+        )
+        self._cql_runner.object_reference_add_concurrent(to_add)
+        return {"object_reference:add": len(to_add)}
+
+    #########################
+    # Deletion
+    #########################
+
+    def _content_delete(self, object_ids: Set[Sha1Git]) -> Dict[str, int]:
+        # We have take care of both Content and SkippedContent and
+        # we don’t know where are each SWHID…
+        # Sadly, we have to read before write which is considered an
+        # anti-pattern for Cassandra. But we only have a SWHID, so a `sha1_git`.
+        # But we need to use the right primary keys to perform a deletion. We
+        # also cannot use `token()` in a WHERE clause of a DELETE… So we have to
+        # perform two SELECTs (in `_content_get_from_hashes()``) to get enough
+        # data to perform our DELETE.
+        content_hashes = []
+        for row in self._content_get_from_hashes("sha1_git", list(object_ids)):
+            content_hashes.append(
+                TotalHashDict(
+                    sha1=row.sha1,
+                    sha1_git=row.sha1_git,
+                    sha256=row.sha256,
+                    blake2s256=row.blake2s256,
+                )
+            )
+        for content_hash in content_hashes:
+            self._cql_runner.content_delete(content_hash)
+        object_ids -= {h["sha1_git"] for h in content_hashes}
+        skipped_content_hashes = []
+        for object_id in object_ids:
+            for token in self._cql_runner.skipped_content_get_tokens_from_single_hash(
+                "sha1_git", object_id
+            ):
+                for row in self._cql_runner.skipped_content_get_from_token(token):
+                    skipped_content_hashes.append(
+                        TotalHashDict(
+                            sha1=row.sha1,
+                            sha1_git=row.sha1_git,
+                            sha256=row.sha256,
+                            blake2s256=row.blake2s256,
+                        )
+                    )
+        for skipped_content_hash in skipped_content_hashes:
+            self._cql_runner.skipped_content_delete(skipped_content_hash)
+        return {
+            "content:delete": len(content_hashes),
+            "content:delete:bytes": 0,
+            "skipped_content:delete": len(skipped_content_hashes),
+        }
+
+    def _directory_delete(self, object_ids: Set[Sha1Git]) -> Dict[str, int]:
+        for object_id in object_ids:
+            self._cql_runner.directory_delete(object_id)
+            self._cql_runner.directory_entry_delete(object_id)
+        # Cassandra makes it so we can’t get the number of affected rows
+        return {"directory:delete": len(object_ids)}
+
+    def _revision_delete(self, object_ids: Set[Sha1Git]) -> Dict[str, int]:
+        for object_id in object_ids:
+            self._cql_runner.revision_delete(object_id)
+            self._cql_runner.revision_parent_delete(object_id)
+        # Cassandra makes it so we can’t get the number of affected rows
+        return {"revision:delete": len(object_ids)}
+
+    def _release_delete(self, object_ids: Set[Sha1Git]) -> Dict[str, int]:
+        for object_id in object_ids:
+            self._cql_runner.release_delete(object_id)
+        # Cassandra makes it so we can’t get the number of affected rows
+        return {"release:delete": len(object_ids)}
+
+    def _snapshot_delete(self, object_ids: Set[Sha1Git]) -> Dict[str, int]:
+        for object_id in object_ids:
+            self._cql_runner.snapshot_delete(object_id)
+            self._cql_runner.snapshot_branch_delete(object_id)
+        # Cassandra makes it so we can’t get the number of affected rows
+        return {"snapshot:delete": len(object_ids)}
+
+    def _origin_delete(self, object_ids: Set[Sha1Git]) -> Dict[str, int]:
+        origin_count = 0
+        origin_visit_count = 0
+        origin_visit_status_count = 0
+        for object_id in object_ids:
+            for origin_row in self._cql_runner.origin_get_by_sha1(object_id):
+                origin_count += 1
+                self._cql_runner.origin_delete(origin_row.sha1)
+                # XXX: We could avoid needless queries if we modify ObjectDeletionInterface
+                for origin_visit_row in self._cql_runner.origin_visit_iter_all(
+                    origin_row.url
+                ):
+                    origin_visit_count += 1
+                    origin_visit_status_count += len(
+                        list(
+                            self._cql_runner.origin_visit_status_get_range(
+                                origin_row.url,
+                                origin_visit_row.visit,
+                                None,
+                                99999999,
+                                ListOrder.ASC,
+                            )
+                        )
+                    )
+                self._cql_runner.origin_visit_delete(origin_row.url)
+                self._cql_runner.origin_visit_status_delete(origin_row.url)
+        return {
+            "origin:delete": origin_count,
+            "origin_visit:delete": origin_visit_count,
+            "origin_visit_status:delete": origin_visit_status_count,
+        }
+
+    def object_delete(self, swhids: List[ExtendedSWHID]) -> Dict[str, int]:
+        """Delete objects from the storage
+
+        All skipped content objects matching the given SWHID will be removed,
+        including those who have the same SWHID due to hash collisions.
+
+        Origin objects are removed alongside their associated origin visit and
+        origin visit status objects.
+
+        Args:
+            swhids: list of SWHID of the objects to remove
+
+        Returns:
+            Summary dict with the following keys and associated values:
+
+                content:delete: Number of content objects removed
+                content:delete:bytes: Sum of the removed contents’ data length
+                skipped_content:delete: Number of skipped content objects removed
+                directory:delete: Number of directory objects removed
+                revision:delete: Number of revision objects removed
+                release:delete: Number of release objects removed
+                snapshot:delete: Number of snapshot objects removed
+                origin:delete: Number of origin objects removed
+                origin_visit:delete: Number of origin visit objects removed
+                origin_visit_status:delete: Number of origin visit status objects removed
+        """
+
+        # groupby() splits consecutive groups, so we need to order the list first
+        def key(swhid: ExtendedSWHID) -> int:
+            return _DELETE_ORDERING[swhid.object_type]
+
+        result = {}
+        sorted_swhids = sorted(swhids, key=key)
+        for object_type, grouped_swhids in itertools.groupby(
+            sorted_swhids, key=operator.attrgetter("object_type")
+        ):
+            object_ids = {swhid.object_id for swhid in grouped_swhids}
+            result.update(_DELETE_METHODS[object_type](self, object_ids))
+        return result
+
     ##########################
     # misc.
     ##########################
@@ -1943,3 +2236,20 @@ class CassandraStorage:
 
     def flush(self, object_types: Sequence[str] = ()) -> Dict[str, int]:
         return {}
+
+
+_DELETE_METHODS: Dict[
+    ExtendedObjectType,
+    Callable[[CassandraStorage, Set[Sha1Git]], Dict[str, int]],
+] = {
+    ExtendedObjectType.CONTENT: CassandraStorage._content_delete,
+    ExtendedObjectType.DIRECTORY: CassandraStorage._directory_delete,
+    ExtendedObjectType.REVISION: CassandraStorage._revision_delete,
+    ExtendedObjectType.RELEASE: CassandraStorage._release_delete,
+    ExtendedObjectType.SNAPSHOT: CassandraStorage._snapshot_delete,
+    ExtendedObjectType.ORIGIN: CassandraStorage._origin_delete,
+}
+
+_DELETE_ORDERING: Dict[ExtendedObjectType, int] = {
+    object_type: order for order, object_type in enumerate(_DELETE_METHODS.keys())
+}
