@@ -7,7 +7,7 @@
 # control
 import logging
 import os
-from typing import Dict, Optional
+from typing import Dict, Optional, Tuple
 
 import click
 
@@ -65,6 +65,32 @@ def storage(ctx, config_file, check_config):
     ctx.obj["check_config"] = check_config
 
 
+@storage.command(name="create-keyspace")
+@click.pass_context
+def create_keyspace(ctx):
+    """Creates a Cassandra keyspace with table definitions suitable for use
+    by swh-storage's Cassandra backend"""
+    from swh.storage.cassandra import create_keyspace
+
+    config = ctx.obj["config"]["storage"]
+
+    for key in ("cls", "hosts", "keyspace", "auth_provider"):
+        if key not in config:
+            ctx.fail(f"Missing {key} key in config file.")
+
+    if config["cls"] != "cassandra":
+        ctx.fail(f"cls must be 'cassandra', not '{config['cls']}'")
+
+    create_keyspace(
+        hosts=config["hosts"],
+        port=config.get("port", 9042),
+        keyspace=config["keyspace"],
+        auth_provider=config["auth_provider"],
+    )
+
+    print("Done.")
+
+
 @storage.command(name="rpc-serve")
 @click.option(
     "--host",
@@ -101,7 +127,7 @@ def serve(ctx, host, port, debug):
     app.run(host, port=int(port), debug=bool(debug))
 
 
-@storage.command()
+@storage.command(name="backfill")
 @click.argument("object_type")
 @click.option("--start-object", default=None)
 @click.option("--end-object", default=None)
@@ -154,7 +180,7 @@ def backfill(ctx, object_type, start_object, end_object, dry_run):
         ctx.exit(0)
 
 
-@storage.command()
+@storage.command(name="replay")
 @click.option(
     "--stop-after-objects",
     "-n",
@@ -189,8 +215,19 @@ def backfill(ctx, object_type, start_object, end_object, dry_run):
     help="Object types to replay",
     multiple=True,
 )
+@click.option(
+    "--known-mismatched-hashes",
+    "-X",
+    "invalid_hashes_file",
+    default=None,
+    type=click.File("r"),
+    help=(
+        "File of SWHIDs of objects that are known to have invalid hashes "
+        "but still need to be replayed."
+    ),
+)
 @click.pass_context
-def replay(ctx, stop_after_objects, object_types):
+def replay(ctx, stop_after_objects, object_types, invalid_hashes_file):
     """Fill a Storage by reading a Journal.
 
     This is typically used for a mirror configuration, reading the Software
@@ -228,6 +265,7 @@ def replay(ctx, stop_after_objects, object_types):
     import functools
 
     from swh.journal.client import get_journal_client
+    from swh.model.swhids import CoreSWHID
     from swh.storage import get_storage
     from swh.storage.replay import ModelObjectDeserializer, process_replay_objects
 
@@ -253,7 +291,26 @@ def replay(ctx, stop_after_objects, object_types):
             "'privileged' is False; we cannot validate anonymized objects."
         )
 
-    deserializer = ModelObjectDeserializer(reporter=reporter, validate=validate)
+    known_mismatched_hashes: Optional[Tuple[Tuple[str, bytes]]] = None
+    if validate and invalid_hashes_file:
+        inv_obj_ids = (
+            row.strip().split(",") for row in invalid_hashes_file if row.strip()
+        )
+        swhids = (
+            (CoreSWHID.from_string(swhid.strip()), bytes.fromhex(computed_id.strip()))
+            for swhid, computed_id in inv_obj_ids
+        )
+
+        known_mismatched_hashes = tuple(
+            (swhid.object_type.name.lower(), swhid.object_id, computed_id)
+            for swhid, computed_id in swhids
+        )
+
+    deserializer = ModelObjectDeserializer(
+        reporter=reporter,
+        validate=validate,
+        known_mismatched_hashes=known_mismatched_hashes,
+    )
 
     client_cfg["value_deserializer"] = deserializer.convert
     if object_types:
@@ -283,6 +340,120 @@ def replay(ctx, stop_after_objects, object_types):
         client.close()
 
 
+@storage.command(name="create-object-reference-partitions")
+@click.argument("start")
+@click.argument("end")
+@click.pass_context
+def create_object_reference_partitions(ctx: click.Context, start: str, end: str):
+    """Create object_reference partitions from START_DATE to END_DATE"""
+
+    import datetime
+
+    from swh.storage import get_storage
+    from swh.storage.postgresql.storage import Storage as PostgreSQLStorage
+
+    storage = get_storage(**ctx.obj["config"]["storage"])
+
+    # This function uses the PostgreSQL swh.storage.db attribute directly, so we
+    # need unwrap all the layers of proxies (e.g. record_references, buffer,
+    # filter, etc.) to get access to the "concrete" underlying storage.
+    while hasattr(storage, "storage"):
+        storage = storage.storage
+
+    if not isinstance(storage, PostgreSQLStorage):
+        ctx.fail("Storage instance needs to be a direct PostgreSQL storage")
+
+    start_date = datetime.date.fromisoformat(start)
+    end_date = datetime.date.fromisoformat(end)
+
+    if start_date > end_date:
+        ctx.fail("Start date must be before end date.")
+
+    cur_date = start_date
+    end_year, end_week = end_date.isocalendar()[0:2]
+    while True:
+        year, week = cur_date.isocalendar()[0:2]
+        if (year, week) > (end_year, end_week):
+            break
+
+        with storage.db() as db:
+            with db.transaction() as cur:
+                monday, sunday = db.object_references_create_partition(
+                    year, week, cur=cur
+                )
+
+        click.echo(
+            "Created object references table for dates between %s and %s"
+            % (monday, sunday)
+        )
+
+        cur_date += datetime.timedelta(days=7)
+
+
+@storage.command(name="remove-old-object-reference-partitions")
+@click.option(
+    "--force",
+    is_flag=True,
+    default=False,
+    help="do not ask for confirmation before removing tables",
+)
+@click.argument("before")
+@click.pass_context
+def remove_old_object_reference_partitions(
+    ctx: click.Context, force: bool, before: str
+):
+    """Remove object_reference partitions for values older than BEFORE"""
+    import datetime
+
+    from swh.storage import get_storage
+    from swh.storage.postgresql.storage import Storage as PostgreSQLStorage
+
+    storage = get_storage(**ctx.obj["config"]["storage"])
+
+    # This function uses the PostgreSQL swh.storage.db attribute directly, so we
+    # need unwrap all the layers of proxies (e.g. record_references, buffer,
+    # filter, etc.) to get access to the "concrete" underlying storage.
+    while hasattr(storage, "storage"):
+        storage = storage.storage
+
+    if not isinstance(storage, PostgreSQLStorage):
+        ctx.fail("Storage instance needs to be a direct PostgreSQL storage")
+
+    before_date = datetime.datetime.fromisoformat(before)
+
+    with storage.db() as db:
+        existing_partitions = db.object_references_list_partitions()
+        to_remove = []
+        for partition in existing_partitions:
+            # If it ends after the specified date, we should keep this partition
+            # and the ones after.
+            if partition.end > before_date:
+                break
+            to_remove.append((partition.year, partition.week))
+
+        if len(to_remove) == 0:
+            click.echo("Nothing needs to be removed.")
+            ctx.exit(0)
+
+        if len(existing_partitions) == len(to_remove):
+            ctx.fail(
+                "Trying to remove all existing partitions. This can’t be right, sorry."
+            )
+
+        if not force:
+            click.echo(
+                "We will remove the following partitions before "
+                f"{before_date.strftime('%Y-%m-%d')}:"
+            )
+            for year, week in to_remove:
+                click.echo(f"- {year:04d} week {week:02d}")
+            click.confirm("Do you want to proceed?", default=False, abort=True)
+
+        with db.transaction() as cur:
+            for year, week in to_remove:
+                db.object_references_drop_partition(year, week, cur)
+
+
 def ensure_check_config(storage_cfg: Dict, check_config: Optional[str], default: str):
     """Helper function to inject the setting of check_config option in the storage config
     dict according to the expected default value (default value depends on the command,
