@@ -55,6 +55,25 @@ class PartialBranches(TypedDict):
     the snapshot has less than the request number of branches."""
 
 
+@attr.s
+class SnapshotBranchByNameResponse:
+    """Object returned by snapshot_branch_get_by_name"""
+
+    branch_found = attr.ib(type=bool)
+    """
+    Branch with the name exists, with or without a target.
+    """
+    target = attr.ib(type=Optional[SnapshotBranch])
+    """
+    Branch target, will be None in case of a dangling branch.
+    """
+    aliases_followed = attr.ib(type=List[bytes])
+    """
+    List of alias names until (including) the target.
+    This will be of length one for all non alias branches.
+    """
+
+
 class HashDict(TypedDict, total=False):
     sha1: bytes
     sha1_git: bytes
@@ -72,6 +91,16 @@ class OriginVisitWithStatuses:
     statuses = attr.ib(type=List[OriginVisitStatus])
 
 
+@attr.s
+class ObjectReference:
+    """Record that the object with SWHID ``source`` references the object with SWHID
+    ``target``, meaning that the ``target`` needs to exist for the ``source`` object
+    to be consistent within the archive."""
+
+    source = attr.ib(type=ExtendedSWHID)
+    target = attr.ib(type=ExtendedSWHID)
+
+
 TResult = TypeVar("TResult")
 PagedResult = CorePagedResult[TResult, str]
 
@@ -364,6 +393,25 @@ class StorageInterface(Protocol):
         """
         ...
 
+    @remote_api_endpoint("content/skipped/find")
+    def skipped_content_find(self, content: HashDict) -> List[SkippedContent]:
+        """Find skipped content for the given hashes
+
+        Args:
+            content: a dictionary representing one content hash, mapping
+                checksum algorithm names (see swh.model.hashutil.ALGORITHMS) to
+                checksum values
+
+        Raises:
+            ValueError: in case the key of the dictionary is not sha1, sha1_git
+                nor sha256.
+
+        Returns:
+            a list of SkippedContent objects matching the search criteria if the
+            skipped content exists. Empty list otherwise.
+        """
+        ...
+
     @remote_api_endpoint("content/skipped/missing")
     def skipped_content_missing(
         self, contents: List[Dict[str, Any]]
@@ -476,13 +524,19 @@ class StorageInterface(Protocol):
         The number of results is not guaranteed to be lower than the ``limit``.
 
         Args:
-            directory_id: dentifier of the directory
+            directory_id: identifier of the directory
             page_token: opaque string used to get the next results of a search
             limit: Number of entries to return
 
         Returns:
             None if the directory does not exist; a page of DirectoryEntry
               objects otherwise.
+
+        See Also:
+            :py:func:`swh.storage.algos.directories.directory_get` will get all
+            entries for a given directory.
+            :py:func:`swh.storage.algos.directories.directory_get_many` will do
+            the same for a set of directories.
         """
         ...
 
@@ -965,14 +1019,12 @@ class StorageInterface(Protocol):
                 contains given prefix
 
         Returns:
-            dict: None if the snapshot does not exist;
-              a dict with three keys otherwise:
-                * **id**: identifier of the snapshot
-                * **branches**: a dict of branches contained in the snapshot
-                  whose keys are the branches' names.
-                * **next_branch**: the name of the first branch not returned
-                  or :const:`None` if the snapshot has less than
-                  `branches_count` branches after `branches_from` included.
+            a PartialBranches object listing a limited amount of branches
+            matching the given criteria or None if the snapshot does not exist.
+
+        See Also:
+            :py:func:`swh.storage.algos.snapshot.snapshot_get_all_branches` will get
+            all branches for a given snapshot.
         """
         ...
 
@@ -985,6 +1037,30 @@ class StorageInterface(Protocol):
         """
         ...
 
+    @remote_api_endpoint("snapshot/branches/get_by_name")
+    def snapshot_branch_get_by_name(
+        self,
+        snapshot_id: Sha1Git,
+        branch_name: bytes,
+        follow_alias_chain: bool = True,
+        max_alias_chain_length: int = 100,
+    ) -> Optional[SnapshotBranchByNameResponse]:
+        """Get a snapshot branch by its name
+
+        Args:
+            snapshot_id: Snapshot identifier
+            branch_name: Branch name to look for
+            follow_alias_chain: If True, find the first non alias branch.
+                Return the first branch (alias or non alias) otherwise
+            max_alias_chain_length: Maximum number of alias chains to be
+                followed before treating the branch as dangling. This has
+                no significance when follow_alias_chain is False.
+
+        Returns:
+            A SnapshotBranchByNameResponse object
+        """
+        ...
+
     @remote_api_endpoint("snapshot/partition/id")
     def snapshot_get_id_partition(
         self,
@@ -1074,12 +1150,15 @@ class StorageInterface(Protocol):
         Returns: Page of OriginVisit data model objects. if next_page_token is None,
             there is no longer data to retrieve.
 
+        See Also:
+            :py:func:`swh.storage.algos.origin.iter_origin_visits` will iterate
+            over all OriginVisits for a given origin.
         """
         ...
 
     @remote_api_endpoint("origin/visit/find_by_date")
     def origin_visit_find_by_date(
-        self, origin: str, visit_date: datetime.datetime
+        self, origin: str, visit_date: datetime.datetime, type: Optional[str] = None
     ) -> Optional[OriginVisit]:
         """Retrieves the origin visit whose date is closest to the provided
         timestamp.
@@ -1088,6 +1167,7 @@ class StorageInterface(Protocol):
         Args:
             origin: origin (URL)
             visit_date: expected visit date
+            type: filter on a specific visit type if provided
 
         Returns:
             A visit if found, None otherwise
@@ -1166,6 +1246,9 @@ class StorageInterface(Protocol):
         Returns: Page of OriginVisitStatus data model objects. if next_page_token is
             None, there is no longer data to retrieve.
 
+        See Also:
+            :py:func:`swh.storage.algos.origin.iter_origin_visit_statuses` will iterate
+            over all OriginVisitStatus objects for a given origin and visit.
         """
         ...
 
@@ -1374,6 +1457,57 @@ class StorageInterface(Protocol):
         """
         ...
 
+    ##########################
+    # reverse index for object references
+    ##########################
+
+    @remote_api_endpoint("object/find_references")
+    def object_find_recent_references(
+        self, target_swhid: ExtendedSWHID, limit: int
+    ) -> List[ExtendedSWHID]:
+        """Return the SWHIDs of objects that are known to reference the object
+        ``target_swhid``.
+
+        Args:
+            target_swhid: the SWHID of the object targeted by the returned objects
+            limit: the maximum number of SWHIDs to return
+
+        Note:
+            The data returned by this function is by essence limited to objects that
+            were recently added to the archive, and is pruned regularly. For
+            completeness, one must also query :mod:`swh.graph` for backwards edges
+            targeting the requested object.
+        """
+        ...
+
+    @remote_api_endpoint("object/references_add")
+    def object_references_add(
+        self, references: List[ObjectReference]
+    ) -> Dict[str, int]:
+        """For each object reference ``(source, target)``, record that the ``source``
+        object references the ``target`` object (meaning that the ``target`` needs to
+        exist for the ``source`` object to be consistent within the archive).
+
+        This function will only be called internally by a reference recording proxy,
+        through one of :func:`directory_add`, :func:`revision_add`, :func:`release_add`,
+        :func:`snapshot_add`, or :func:`origin_visit_status_add`. External users of
+        :mod:`swh.storage` should not need to use this function directly.
+
+        Note:
+            these records are inserted in time-based partitions that can be pruned when
+            the objects are known in an up-to-date :mod:`swh.graph` instance.
+
+        Args:
+            references: a list of ``(source, target)`` SWHID tuples
+
+        Returns:
+            A summary dict with the following keys:
+
+              object_reference:add: the number of object references added
+
+        """
+        ...
+
     ##########################
     # misc.
     ##########################
@@ -1567,3 +1701,37 @@ class StorageInterface(Protocol):
         actual writes to the backend.
         """
         ...
+
+
+class ObjectDeletionInterface(Protocol):
+    def object_delete(self, swhids: List[ExtendedSWHID]):
+        """Delete objects from the storage
+
+        All skipped content objects matching the given SWHID will be removed,
+        including those who have the same SWHID due to hash collisions.
+
+        Origin objects are removed alongside their associated origin visit and
+        origin visit status objects.
+
+        Only objects from this facility will be removed. The same method
+        should be called on other storage, objstorage, or journal instances
+        where the specified objects need to be removed.
+
+        Args:
+            swhids: list of SWHID of the objects to remove
+
+        Returns:
+            Summary dict with the following keys and associated values:
+
+                content:delete: Number of content objects removed
+                content:delete:bytes: Sum of the removed contents’ data length
+                skipped_content:delete: Number of skipped content objects removed
+                directory:delete: Number of directory objects removed
+                revision:delete: Number of revision objects removed
+                release:delete: Number of release objects removed
+                snapshot:delete: Number of snapshot objects removed
+                origin:delete: Number of origin objects removed
+                origin_visit:delete: Number of origin visit objects removed
+                origin_visit_status:delete: Number of origin visit status objects removed
+        """
+        ...
