@@ -5,7 +5,7 @@
 
 import datetime
 import math
-from typing import Any, Dict, Optional
+from typing import Any, Dict, Optional, Tuple
 import warnings
 
 from swh.core.utils import encode_with_unescape
@@ -24,8 +24,11 @@ from swh.model.model import (
     Timestamp,
     TimestampWithTimezone,
 )
-from swh.model.swhids import CoreSWHID, ExtendedSWHID
+from swh.model.swhids import CoreSWHID
+from swh.model.swhids import ExtendedObjectType as SwhidExtendedObjectType
+from swh.model.swhids import ExtendedSWHID
 from swh.model.swhids import ObjectType as SwhidObjectType
+from swh.storage.interface import ObjectReference
 
 from ..utils import map_optional
 
@@ -249,13 +252,10 @@ def db_to_revision(db_revision: Dict[str, Any]) -> Revision:
                 parents.append(parent)
 
     metadata = db_revision["metadata"]
-    extra_headers = db_revision["extra_headers"]
-    if not extra_headers:
-        if metadata and "extra_headers" in metadata:
-            extra_headers = db_to_git_headers(metadata.pop("extra_headers"))
-        else:
-            # For older versions of the database that were not migrated to schema v161
-            extra_headers = ()
+    if metadata and "extra_headers" in metadata:
+        del metadata["extra_headers"]
+
+    extra_headers = db_revision["extra_headers"] or ()
 
     return Revision(
         id=db_revision["id"],
@@ -378,3 +378,21 @@ def db_to_extid(row) -> ExtID:
             object_type=SwhidObjectType[row["target_type"].upper()],
         ),
     )
+
+
+def object_reference_to_db(
+    reference: ObjectReference,
+) -> Tuple[str, bytes, str, bytes]:
+    return (
+        reference.source.object_type.name.lower(),
+        reference.source.object_id,
+        reference.target.object_type.name.lower(),
+        reference.target.object_id,
+    )
+
+
+def db_to_object_reference_source(row) -> ExtendedSWHID:
+    return ExtendedSWHID(
+        object_id=row["source"],
+        object_type=SwhidExtendedObjectType[row["source_type"].upper()],
+    )
