@@ -8,6 +8,7 @@ from collections import defaultdict
 import contextlib
 from contextlib import contextmanager
 import datetime
+import functools
 import itertools
 import logging
 import operator
@@ -32,7 +33,8 @@ import psycopg2.errors
 import psycopg2.pool
 
 from swh.core.api.serializers import msgpack_dumps, msgpack_loads
-from swh.core.db.common import db_transaction, db_transaction_generator
+from swh.core.db.common import db_transaction as _db_transaction
+from swh.core.db.common import db_transaction_generator as _db_transaction_generator
 from swh.core.db.db_utils import swh_db_flavor, swh_db_version
 from swh.model.hashutil import DEFAULT_ALGORITHMS, hash_to_bytes, hash_to_hex
 from swh.model.model import (
@@ -60,6 +62,7 @@ from swh.model.model import (
 from swh.model.swhids import ExtendedObjectType, ExtendedSWHID, ObjectType
 from swh.storage.exc import (
     HashCollision,
+    QueryTimeout,
     StorageArgumentException,
     StorageDBError,
     UnknownMetadataAuthority,
@@ -69,9 +72,11 @@ from swh.storage.interface import (
     VISIT_STATUSES,
     HashDict,
     ListOrder,
+    ObjectReference,
     OriginVisitWithStatuses,
     PagedResult,
     PartialBranches,
+    SnapshotBranchByNameResponse,
 )
 from swh.storage.objstorage import ObjStorage
 from swh.storage.utils import (
@@ -125,6 +130,38 @@ def convert_validation_exceptions():
         raise StorageArgumentException(str(e))
 
 
+def db_transaction_generator(*args, **kwargs):
+    def decorator(meth):
+        meth = _db_transaction_generator(*args, **kwargs)(meth)
+
+        @functools.wraps(meth)
+        def _meth(self, *args, **kwargs):
+            try:
+                yield from meth(self, *args, **kwargs)
+            except psycopg2.errors.QueryCanceled as e:
+                raise QueryTimeout(*e.args)
+
+        return _meth
+
+    return decorator
+
+
+def db_transaction(*args, **kwargs):
+    def decorator(meth):
+        meth = _db_transaction(*args, **kwargs)(meth)
+
+        @functools.wraps(meth)
+        def _meth(self, *args, **kwargs):
+            try:
+                return meth(self, *args, **kwargs)
+            except psycopg2.errors.QueryCanceled as e:
+                raise QueryTimeout(*e.args)
+
+        return _meth
+
+    return decorator
+
+
 TRow = TypeVar("TRow", bound=Tuple)
 TResult = TypeVar("TResult")
 
@@ -179,12 +216,12 @@ def _get_paginated_sha1_partition(
 class Storage:
     """SWH storage datastore proxy, encompassing DB and object storage"""
 
-    current_version: int = 186
+    current_version: int = 192
 
     def __init__(
         self,
         db,
-        objstorage,
+        objstorage=None,
         min_pool_conns=1,
         max_pool_conns=10,
         journal_writer=None,
@@ -200,7 +237,8 @@ class Storage:
 
         Args:
             db: either a libpq connection string, or a psycopg2 connection
-            objstorage: configuration for the backend :class:`ObjStorage`
+            objstorage: configuration for the backend :class:`ObjStorage`; if unset,
+               use a NoopObjStorage
             min_pool_conns: min number of connections in the psycopg2 pool
             max_pool_conns: max number of connections in the psycopg2 pool
             journal_writer: configuration for the :class:`JournalWriter`
@@ -285,7 +323,6 @@ class Storage:
 
     @db_transaction()
     def check_config(self, *, check_write: bool, db: Db, cur=None) -> bool:
-
         if not self.objstorage.check_config(check_write=check_write):
             return False
 
@@ -369,9 +406,9 @@ class Storage:
 
         with self.db() as db:
             with db.transaction() as cur:
-                missing = list(
+                missing = set(
                     self.content_missing(
-                        map(Content.to_dict, contents),
+                        [c.to_dict() for c in contents],
                         key_hash="sha1_git",
                         db=db,
                         cur=cur,
@@ -405,11 +442,13 @@ class Storage:
     def content_add_metadata(
         self, content: List[Content], *, db: Db, cur=None
     ) -> Dict[str, int]:
-        missing = self.content_missing(
-            (c.to_dict() for c in content),
-            key_hash="sha1_git",
-            db=db,
-            cur=cur,
+        missing = set(
+            self.content_missing(
+                [c.to_dict() for c in content],
+                key_hash="sha1_git",
+                db=db,
+                cur=cur,
+            )
         )
         contents = [c for c in content if c.sha1_git in missing]
 
@@ -589,13 +628,40 @@ class Storage:
             "skipped_content:add": len(content),
         }
 
+    @db_transaction()
+    def skipped_content_find(
+        self, content: HashDict, *, db: Db, cur=None
+    ) -> List[SkippedContent]:
+        if not set(content).intersection(DEFAULT_ALGORITHMS):
+            raise StorageArgumentException(
+                "content keys must contain at least one "
+                f"of: {', '.join(sorted(DEFAULT_ALGORITHMS))}"
+            )
+
+        rows = db.skipped_content_find(
+            sha1=content.get("sha1"),
+            sha1_git=content.get("sha1_git"),
+            sha256=content.get("sha256"),
+            blake2s256=content.get("blake2s256"),
+            cur=cur,
+        )
+        skipped_contents = []
+        for row in rows:
+            row_d = dict(zip(db.skipped_content_find_cols, row))
+            skipped_contents.append(SkippedContent(**row_d))
+        return skipped_contents
+
     @db_transaction_generator()
     def skipped_content_missing(
         self, contents: List[Dict[str, Any]], *, db: Db, cur=None
     ) -> Iterable[Dict[str, Any]]:
         contents = list(contents)
         for content in db.skipped_content_missing(contents, cur):
-            yield dict(zip(db.content_hash_keys, content))
+            yield {
+                algo: hash
+                for (algo, hash) in zip(db.content_hash_keys, content)
+                if hash
+            }
 
     ##########################
     # Directory
@@ -1150,7 +1216,6 @@ class Storage:
         db: Db,
         cur=None,
     ) -> Optional[PartialBranches]:
-
         if snapshot_id == EMPTY_SNAPSHOT_ID:
             return PartialBranches(
                 id=snapshot_id,
@@ -1207,6 +1272,73 @@ class Storage:
     def snapshot_get_random(self, *, db: Db, cur=None) -> Sha1Git:
         return db.snapshot_get_random(cur)
 
+    @db_transaction(statement_timeout=2000)
+    def snapshot_branch_get_by_name(
+        self,
+        snapshot_id: Sha1Git,
+        branch_name: bytes,
+        db: Db,
+        cur=None,
+        follow_alias_chain: bool = True,
+        max_alias_chain_length: int = 100,
+    ) -> Optional[SnapshotBranchByNameResponse]:
+        if list(self.snapshot_missing([snapshot_id])):
+            return None
+
+        if snapshot_id == EMPTY_SNAPSHOT_ID:
+            return SnapshotBranchByNameResponse(
+                branch_found=False,
+                target=None,
+                aliases_followed=[],
+            )
+
+        cols_to_fetch = ["target", "target_type"]
+        resolve_chain: List[bytes] = []
+        while True:
+            branch = db.snapshot_branch_get_by_name(
+                cols_to_fetch=cols_to_fetch,
+                snapshot_id=snapshot_id,
+                branch_name=branch_name,
+                cur=cur,
+            )
+            if branch is None:
+                # target branch is None, there could be items in aliases_followed
+                target = None
+                break
+            branch_d = dict(zip(cols_to_fetch, branch))
+            resolve_chain.append(branch_name)
+            if (
+                branch_d["target_type"] != TargetType.ALIAS.value
+                or not follow_alias_chain
+            ):
+                # first non alias branch or the first branch when follow_alias_chain is False
+                target = (
+                    SnapshotBranch(
+                        target=branch_d["target"],
+                        target_type=TargetType(branch_d["target_type"]),
+                    )
+                    if branch_d["target"]
+                    else None
+                )
+                break
+            elif (
+                # Circular reference
+                resolve_chain.count(branch_name) > 1
+                # Too many re-directs
+                or len(resolve_chain) >= max_alias_chain_length
+            ):
+                target = None
+                break
+            # Branch has a non-None target with type alias
+            branch_name = branch_d["target"]
+
+        return SnapshotBranchByNameResponse(
+            # resolve_chian has items, brach_found must be True
+            branch_found=bool(resolve_chain),
+            target=target,
+            aliases_followed=resolve_chain,
+        )
+
     ##########################
     # OriginVisit and OriginVisitStatus
     ##########################
@@ -1392,7 +1524,6 @@ class Storage:
         next_page_token = visits_page.next_page_token
 
         if visits:
-
             visit_from = min(visits[0].visit, visits[-1].visit)
             visit_to = max(visits[0].visit, visits[-1].visit)
 
@@ -1419,9 +1550,15 @@ class Storage:
 
     @db_transaction(statement_timeout=2000)
     def origin_visit_find_by_date(
-        self, origin: str, visit_date: datetime.datetime, *, db: Db, cur=None
+        self,
+        origin: str,
+        visit_date: datetime.datetime,
+        type: Optional[str] = None,
+        *,
+        db: Db,
+        cur=None,
     ) -> Optional[OriginVisit]:
-        row_d = db.origin_visit_find_by_date(origin, visit_date, cur=cur)
+        row_d = db.origin_visit_find_by_date(origin, visit_date, type=type, cur=cur)
         if not row_d:
             return None
         return OriginVisit(
@@ -1888,6 +2025,36 @@ class Storage:
             return None
         return MetadataAuthority.from_dict(dict(zip(db.metadata_authority_cols, row)))
 
+    #########################
+    # 'object_references' table
+    #########################
+
+    @db_transaction()
+    def object_find_recent_references(
+        self, target_swhid: ExtendedSWHID, limit: int, *, db: Db, cur=None
+    ) -> List[ExtendedSWHID]:
+        return [
+            converters.db_to_object_reference_source(row)
+            for row in db.object_references_get(
+                target_type=target_swhid.object_type.name.lower(),
+                target=target_swhid.object_id,
+                limit=limit,
+                cur=cur,
+            )
+        ]
+
+    @db_transaction()
+    def object_references_add(
+        self, references: List[ObjectReference], *, db: Db, cur=None
+    ) -> Dict[str, int]:
+        to_add = list({converters.object_reference_to_db(ref) for ref in references})
+        db.object_references_add(
+            to_add,
+            cur=cur,
+        )
+
+        return {"object_reference:add": len(to_add)}
+
     def clear_buffers(self, object_types: Sequence[str] = ()) -> None:
         """Do nothing"""
         return None
@@ -1908,3 +2075,41 @@ class Storage:
         if not fetcher_id:
             raise UnknownMetadataFetcher(str(fetcher))
         return fetcher_id
+
+    #########################
+    # ObjectDeletionInterface
+    #########################
+
+    @db_transaction()
+    def object_delete(
+        self, swhids: List[ExtendedSWHID], *, db: Db, cur=None
+    ) -> Dict[str, int]:
+        """Delete objects from the storage
+
+        All skipped content objects matching the given SWHID will be removed,
+        including those who have the same SWHID due to hash collisions.
+
+        Origin objects are removed alongside their associated origin visit and
+        origin visit status objects.
+
+        Args:
+            swhids: list of SWHID of the objects to remove
+
+        Returns:
+            Summary dict with the following keys and associated values:
+
+                content:delete: Number of content objects removed
+                content:delete:bytes: Sum of the removed contents’ data length
+                skipped_content:delete: Number of skipped content objects removed
+                directory:delete: Number of directory objects removed
+                revision:delete: Number of revision objects removed
+                release:delete: Number of release objects removed
+                snapshot:delete: Number of snapshot objects removed
+                origin:delete: Number of origin objects removed
+                origin_visit:delete: Number of origin visit objects removed
+                origin_visit_status:delete: Number of origin visit status objects removed
+        """
+        object_rows = [
+            (swhid.object_type.name.lower(), swhid.object_id) for swhid in swhids
+        ]
+        return db.object_delete(object_rows)
