@@ -1,3 +1,8 @@
+select swh_get_dbflavor() != 'only_masking' as dbflavor_not_only_masking \gset
+
+-- This skips this whole file if the dbflavor is `only_masking`
+\if :dbflavor_not_only_masking
+
 create or replace function hash_sha1(text)
        returns text
 as $$
@@ -224,18 +229,22 @@ begin
     select distinct t.target, t.name, t.perms
     from tmp_directory_entry_%1$s t
     where not exists (
-    select 1
-    from directory_entry_%1$s i
-    where t.target = i.target and t.name = i.name and t.perms = i.perms)
+      select 1
+      from directory_entry_%1$s i
+      where t.target = i.target and t.name = i.name and t.perms = i.perms
+	)
+	on conflict
+	do nothing
+	;
    ', typ);
 
     execute format('
     with new_entries as (
-	select t.dir_id, array_agg(i.id) as entries
-	from tmp_directory_entry_%1$s t
-	inner join directory_entry_%1$s i
-	using (target, name, perms)
-	group by t.dir_id
+      select t.dir_id, array_agg(i.id) as entries
+      from tmp_directory_entry_%1$s t
+      inner join directory_entry_%1$s i
+      using (target, name, perms)
+      group by t.dir_id
     )
     update tmp_directory as d
     set %1$s_entries = new_entries.entries
@@ -268,7 +277,8 @@ begin
     select id, dir_entries, file_entries, rev_entries, raw_manifest from tmp_directory t
     where not exists (
         select 1 from directory d
-	where d.id = t.id);
+        where d.id = t.id
+	);
 
     return;
 end
@@ -815,7 +825,7 @@ $$;
 
 -- Find the visit of origin closest to date visit_date
 -- Breaks ties by selecting the largest visit id
-create or replace function swh_visit_find_by_date(origin_url text, visit_date timestamptz default NOW())
+create or replace function swh_visit_find_by_date(origin_url text, visit_date timestamptz default NOW(), visit_type text default null)
     returns setof origin_visit
     language plpgsql
     stable
@@ -825,21 +835,34 @@ declare
 begin
   select id into origin_id from origin where url=origin_url;
   return query
-  with closest_two_visits as ((
-    select ov, (date - visit_date) as interval, visit
+  -- first find the two closest dates (this does a one-row scan of the origin_visit
+  -- (origin, date) index twice, forward and backward around the given visit_date)
+  with closest_two_visit_dates as ((
+    select date, (date - visit_date) as interval -- date >= visit_date so interval >= 0
     from origin_visit ov
     where ov.origin = origin_id
           and ov.date >= visit_date
-    order by ov.date asc, ov.visit desc
+          and (visit_type is null or ov.type = visit_type)
+    order by ov.date asc
     limit 1
   ) union (
-    select ov, (visit_date - date) as interval, visit
+    select date, (visit_date - date) as interval -- date < visit_date so interval > 0
     from origin_visit ov
     where ov.origin = origin_id
           and ov.date < visit_date
-    order by ov.date desc, ov.visit desc
+          and (visit_type is null or ov.type = visit_type)
+    order by ov.date desc
     limit 1
-  )) select (ov).* from closest_two_visits order by interval, visit limit 1;
+  ))
+  -- then select the data of the visit at the closest date, with the highest visit id
+  -- (this uses the origin_visit (origin, date) index a third time to find the highest
+  -- visit id at the given date)
+  select * from origin_visit
+    where origin = origin_id
+      and date = (select date from closest_two_visit_dates order by interval limit 1)
+      and (visit_type is null or type = visit_type)
+    order by visit desc
+    limit 1;
 end
 $$;
 
@@ -1029,3 +1052,6 @@ create trigger update_counts_from_bucketed
   for each row
   when (NEW.line % 256 = 0)
   execute procedure swh_update_counters_from_buckets();
+
+-- :dbflavor_not_only_masking
+\endif
