@@ -1,17 +1,13 @@
-# Copyright (C) 2018-2021  The Software Heritage developers
+# Copyright (C) 2018-2023  The Software Heritage developers
 # See the AUTHORS file at the top-level directory of this distribution
 # License: GNU General Public License version 3, or any later version
 # See top-level LICENSE file for more information
 
+import dataclasses
 import datetime
 import itertools
-import os
-import resource
-import signal
-import socket
-import subprocess
-import time
-from typing import Any, Dict
+import re
+from typing import Any, Dict, Union
 
 import attr
 from cassandra.cluster import NoHostAvailable
@@ -19,12 +15,19 @@ import pytest
 
 from swh.core.api.classes import stream_results
 from swh.model import from_disk
-from swh.model.model import Directory, DirectoryEntry, Snapshot, SnapshotBranch
+from swh.model.model import (
+    Directory,
+    DirectoryEntry,
+    Person,
+    Snapshot,
+    SnapshotBranch,
+    TimestampWithTimezone,
+)
 from swh.storage import get_storage
-from swh.storage.cassandra import create_keyspace
 from swh.storage.cassandra.cql import BATCH_INSERT_MAX_SIZE
-from swh.storage.cassandra.model import ContentRow, ExtIDRow
-from swh.storage.cassandra.schema import HASH_ALGORITHMS, TABLES
+import swh.storage.cassandra.model
+from swh.storage.cassandra.model import BaseRow, ContentRow, ExtIDRow
+from swh.storage.cassandra.schema import CREATE_TABLES_QUERIES, HASH_ALGORITHMS, TABLES
 from swh.storage.cassandra.storage import DIRECTORY_ENTRIES_INSERT_ALGOS
 from swh.storage.tests.storage_data import StorageData
 from swh.storage.tests.storage_tests import (
@@ -33,185 +36,117 @@ from swh.storage.tests.storage_tests import (
 from swh.storage.tests.storage_tests import TestStorage as _TestStorage
 from swh.storage.utils import now, remove_keys
 
-CONFIG_TEMPLATE = """
-data_file_directories:
-    - {data_dir}/data
-commitlog_directory: {data_dir}/commitlog
-hints_directory: {data_dir}/hints
-saved_caches_directory: {data_dir}/saved_caches
 
-commitlog_sync: periodic
-commitlog_sync_period_in_ms: 1000000
-partitioner: org.apache.cassandra.dht.Murmur3Partitioner
-endpoint_snitch: SimpleSnitch
-seed_provider:
-    - class_name: org.apache.cassandra.locator.SimpleSeedProvider
-      parameters:
-          - seeds: "127.0.0.1"
-
-storage_port: {storage_port}
-native_transport_port: {native_transport_port}
-start_native_transport: true
-listen_address: 127.0.0.1
-
-enable_user_defined_functions: true
-
-# speed-up by disabling period saving to disk
-key_cache_save_period: 0
-row_cache_save_period: 0
-trickle_fsync: false
-commitlog_sync_period_in_ms: 100000
-"""
-
-SCYLLA_EXTRA_CONFIG_TEMPLATE = """
-experimental_features:
-    - udf
-view_hints_directory: {data_dir}/view_hints
-prometheus_port: 0  # disable prometheus server
-start_rpc: false  # disable thrift server
-api_port: {api_port}
-"""
+@pytest.fixture
+def swh_storage_backend_config(swh_storage_cassandra_backend_config):
+    return swh_storage_cassandra_backend_config
 
 
-def free_port():
-    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-    sock.bind(("127.0.0.1", 0))
-    port = sock.getsockname()[1]
-    sock.close()
-    return port
+def _python_type_to_cql_type_re(ty: type) -> str:
+    if ty is bytes:
+        return "blob"
+    elif ty is str:
+        return "(ascii|text)"
+    elif ty is int:
+        return "(small||big)int"
+    elif ty is datetime.datetime:
+        return "timestamp"
+    elif ty is TimestampWithTimezone:
+        return "microtimestamp_with_timezone"
+    elif ty is Person:
+        return "person"
+    elif ty is bool:
+        return "boolean"
+    elif ty is dict:
+        return "frozen<list <list<blob>> >"
+    elif getattr(ty, "__origin__", None) is Union:
+        if len(ty.__args__) == 2 and type(None) in ty.__args__:  # type: ignore
+            (inner_type,) = [
+                arg
+                for arg in ty.__args__  # type: ignore
+                if arg is not type(None)  # noqa
+            ]
+            # This is Optional[inner_type]. CQL does not support NOT NULL,
+            # so the corresponding type is:
+            return _python_type_to_cql_type_re(inner_type)
+
+    assert False, f"Unsupported type: {ty}"
 
 
-def wait_for_peer(addr, port):
-    wait_until = time.time() + 60
-    while time.time() < wait_until:
-        try:
-            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-            sock.connect((addr, port))
-        except ConnectionRefusedError:
-            time.sleep(0.1)
-        else:
-            sock.close()
-            return True
-    return False
+@pytest.mark.cassandra
+def test_schema_matches_model():
+    """Checks tables defined in :mod:`swh.storage.cassandra.schema` match
+    the object model defined in :mod:`swh.storage.cassandra.model`.
+    """
 
+    models = {}  # {table_name: cls}
+    for obj_name in dir(swh.storage.cassandra.model):
+        cls = getattr(swh.storage.cassandra.model, obj_name)
+        if isinstance(cls, type) and issubclass(cls, BaseRow) and hasattr(cls, "TABLE"):
+            assert cls.TABLE not in models, f"Duplicate model for table {cls.TABLE}"
+            models[cls.TABLE] = cls
 
-@pytest.fixture(scope="session")
-def cassandra_cluster(tmpdir_factory):
-    cassandra_conf = tmpdir_factory.mktemp("cassandra_conf")
-    cassandra_data = tmpdir_factory.mktemp("cassandra_data")
-    cassandra_log = tmpdir_factory.mktemp("cassandra_log")
-    native_transport_port = free_port()
-    storage_port = free_port()
-    jmx_port = free_port()
-    api_port = free_port()
+    del models["object_count"]  # https://forge.softwareheritage.org/D6150
 
-    use_scylla = bool(os.environ.get("SWH_USE_SCYLLADB", ""))
-
-    cassandra_bin = os.environ.get(
-        "SWH_CASSANDRA_BIN", "/usr/bin/scylla" if use_scylla else "/usr/sbin/cassandra"
-    )
-
-    if use_scylla:
-        os.makedirs(cassandra_conf.join("conf"))
-        config_path = cassandra_conf.join("conf/scylla.yaml")
-        config_template = CONFIG_TEMPLATE + SCYLLA_EXTRA_CONFIG_TEMPLATE
-    else:
-        config_path = cassandra_conf.join("cassandra.yaml")
-        config_template = CONFIG_TEMPLATE
-
-    with open(str(config_path), "w") as fd:
-        fd.write(
-            config_template.format(
-                data_dir=str(cassandra_data),
-                storage_port=storage_port,
-                native_transport_port=native_transport_port,
-                api_port=api_port,
+    statements = {}  # {table_name: statement}
+    for statement in CREATE_TABLES_QUERIES:
+        statement = re.sub(" +", " ", statement.strip())  # normalize whitespace
+        if statement.startswith(
+            (
+                "CREATE TYPE ",
+                "CREATE OR REPLACE FUNCTION",
+                "CREATE OR REPLACE AGGREGATE",
+                "-- Secondary table",
             )
-        )
+        ):
+            continue
+        prefix = "CREATE TABLE IF NOT EXISTS "
+        assert statement.startswith(prefix)
+        table_name = statement[len(prefix) :].split(" ", 1)[0]
+        assert (
+            table_name not in statements
+        ), f"Duplicate statement for table {table_name}"
 
-    if os.environ.get("SWH_CASSANDRA_LOG"):
-        stdout = stderr = None
-    else:
-        stdout = stderr = subprocess.DEVNULL
+        statements[table_name] = statement
 
-    env = {
-        "MAX_HEAP_SIZE": "300M",
-        "HEAP_NEWSIZE": "50M",
-        "JVM_OPTS": "-Xlog:gc=error:file=%s/gc.log" % cassandra_log,
-    }
-    if "JAVA_HOME" in os.environ:
-        env["JAVA_HOME"] = os.environ["JAVA_HOME"]
+    assert set(models) - set(statements) == set(), "Missing statements"
+    assert set(statements) - set(models) == set(), "Missing models"
 
-    if use_scylla:
-        env = {
-            **env,
-            "SCYLLA_HOME": cassandra_conf,
-        }
-        # prevent "NOFILE rlimit too low (recommended setting 200000,
-        # minimum setting 10000; refusing to start."
-        resource.setrlimit(resource.RLIMIT_NOFILE, (200000, 200000))
+    mismatches = []
+    for table_name in list(models):
+        # Compute what we expect the statement to look like:
+        model = models[table_name]
+        expected_lines = [rf"CREATE TABLE IF NOT EXISTS {model.TABLE} \("]
+        for field in dataclasses.fields(model):
+            expected_lines.append(
+                f" {field.name} {_python_type_to_cql_type_re(field.type)},"
+            )
 
-        proc = subprocess.Popen(
-            [
-                cassandra_bin,
-                "--developer-mode=1",
-            ],
-            start_new_session=True,
-            env=env,
-            stdout=stdout,
-            stderr=stderr,
-        )
-    else:
-        proc = subprocess.Popen(
-            [
-                cassandra_bin,
-                "-Dcassandra.config=file://%s/cassandra.yaml" % cassandra_conf,
-                "-Dcassandra.logdir=%s" % cassandra_log,
-                "-Dcassandra.jmx.local.port=%d" % jmx_port,
-                "-Dcassandra-foreground=yes",
-            ],
-            start_new_session=True,
-            env=env,
-            stdout=stdout,
-            stderr=stderr,
-        )
+        partition_key = ", ".join(model.PARTITION_KEY)
+        clustering_key = "".join(", " + col for col in model.CLUSTERING_KEY)
+        expected_lines.append(rf" PRIMARY KEY \(\({partition_key}\){clustering_key}\)")
 
-    listening = wait_for_peer("127.0.0.1", native_transport_port)
-
-    if listening:
-        yield (["127.0.0.1"], native_transport_port)
-
-    if not listening or os.environ.get("SWH_CASSANDRA_LOG"):
-        debug_log_path = str(cassandra_log.join("debug.log"))
-        if os.path.exists(debug_log_path):
-            with open(debug_log_path) as fd:
-                print(fd.read())
-
-    if not listening:
-        if proc.poll() is None:
-            raise Exception("cassandra process unexpectedly not listening.")
+        if table_name == "origin_visit_status":
+            # we need to special-case this one
+            expected_lines.append(r"\)")
+            expected_lines.append(r"WITH CLUSTERING ORDER BY \(visit DESC, date DESC\)")
+            expected_lines.append(r";")
         else:
-            raise Exception("cassandra process unexpectedly stopped.")
+            expected_lines.append(r"\);")
 
-    pgrp = os.getpgid(proc.pid)
-    os.killpg(pgrp, signal.SIGKILL)
+        statement = statements[table_name]
+        actual_lines = statement.split("\n")
 
+        mismatches.extend(
+            (table_name, expected_line, actual_line)
+            for (expected_line, actual_line) in zip(expected_lines, actual_lines)
+            if not re.match(expected_line, actual_line)
+        )
 
-class RequestHandler:
-    def on_request(self, rf):
-        if hasattr(rf.message, "query"):
-            print()
-            print(rf.message.query)
-
-
-@pytest.fixture(scope="session")
-def keyspace(cassandra_cluster):
-    (hosts, port) = cassandra_cluster
-    keyspace = "test" + os.urandom(10).hex()
-
-    create_keyspace(hosts, keyspace, port)
-
-    return keyspace
+    assert not mismatches, "\n" + "\n".join(
+        f"{table_name}:\t{actual_line!r} did not match {expected_line!r}"
+        for (table_name, expected_line, actual_line) in mismatches
+    )
 
 
 # tests are executed using imported classes (TestStorage and
@@ -219,29 +154,6 @@ def keyspace(cassandra_cluster):
 # below
 
 
-@pytest.fixture
-def swh_storage_backend_config(cassandra_cluster, keyspace):
-    (hosts, port) = cassandra_cluster
-
-    storage_config = dict(
-        cls="cassandra",
-        hosts=hosts,
-        port=port,
-        keyspace=keyspace,
-        journal_writer={"cls": "memory"},
-        objstorage={"cls": "memory"},
-    )
-
-    yield storage_config
-
-    storage = get_storage(**storage_config)
-
-    for table in TABLES:
-        storage._cql_runner._session.execute(f"TRUNCATE TABLE {keyspace}.{table}")
-
-    storage._cql_runner._cluster.shutdown()
-
-
 @pytest.mark.cassandra
 class TestCassandraStorage(_TestStorage):
     def test_config_wrong_consistency_should_raise(self):
@@ -268,6 +180,22 @@ class TestCassandraStorage(_TestStorage):
         with pytest.raises(NoHostAvailable):
             storage.content_get_random()
 
+    def test_no_authentication(self, swh_storage_backend_config):
+        config_without_authentication = swh_storage_backend_config.copy()
+        config_without_authentication.pop("auth_provider")
+
+        with pytest.raises(NoHostAvailable):
+            get_storage(**config_without_authentication)
+
+    def test_no_authenticator_class(self, swh_storage_backend_config):
+        config_without_authentication = swh_storage_backend_config.copy()
+        authenticator_config = config_without_authentication.pop("auth_provider").copy()
+        authenticator_config.pop("cls")
+        config_without_authentication["auth_provider"] = authenticator_config
+
+        with pytest.raises(ValueError, match="cls property"):
+            get_storage(**config_without_authentication)
+
     def test_content_add_murmur3_collision(self, swh_storage, mocker, sample_data):
         """The Murmur3 token is used as link from index tables to the main
         table; and non-matching contents with colliding murmur3-hash
@@ -691,6 +619,70 @@ class TestCassandraStorage(_TestStorage):
     def test_origin_count(self):
         pass
 
+    def test_object_delete(self, swh_storage, sample_data):
+        # For our sanity checks
+        affected_tables = set(TABLES) - {
+            "raw_extrinsic_metadata",
+            "metadata_authority",
+            "metadata_fetcher",
+            "extid",
+            "extid_by_target",
+            "object_references",
+        }
+        # XXX: Not ideal…
+        cql_runner = getattr(swh_storage, "_cql_runner")
+        execute_query = getattr(cql_runner, "_execute_with_retries")
+
+        swh_storage.content_add(sample_data.contents)
+        swh_storage.skipped_content_add(sample_data.skipped_contents)
+        swh_storage.directory_add(sample_data.directories)
+        swh_storage.revision_add(sample_data.revisions)
+        swh_storage.release_add(sample_data.releases)
+        swh_storage.snapshot_add(sample_data.snapshots)
+        swh_storage.origin_add(sample_data.origins)
+        swh_storage.origin_visit_add(sample_data.origin_visits)
+        swh_storage.origin_visit_status_add(sample_data.origin_visit_statuses)
+        swhids = (
+            [content.swhid().to_extended() for content in sample_data.contents]
+            + [
+                skipped_content.swhid().to_extended()
+                for skipped_content in sample_data.skipped_contents
+            ]
+            + [directory.swhid().to_extended() for directory in sample_data.directories]
+            + [revision.swhid().to_extended() for revision in sample_data.revisions]
+            + [release.swhid().to_extended() for release in sample_data.releases]
+            + [snapshot.swhid().to_extended() for snapshot in sample_data.snapshots]
+            + [origin.swhid() for origin in sample_data.origins]
+        )
+
+        # Do we have something in every affected tables?
+        for table in affected_tables:
+            row = execute_query(
+                f"SELECT COUNT(*) AS count FROM {cql_runner.keyspace}.{table}", []
+            ).one()
+            assert row["count"] >= 1, f"nothing in table {table}"
+
+        result = swh_storage.object_delete(swhids)
+        assert result == {
+            "content:delete": 3,
+            "content:delete:bytes": 0,
+            "skipped_content:delete": 2,
+            "directory:delete": 7,
+            "revision:delete": 8,
+            "release:delete": 3,
+            "snapshot:delete": 3,
+            "origin:delete": 7,
+            "origin_visit:delete": 3,
+            "origin_visit_status:delete": 3,
+        }
+
+        # Have we cleaned every affected tables?
+        for table in affected_tables:
+            row = execute_query(
+                f"SELECT COUNT(*) AS count FROM {cql_runner.keyspace}.{table}", []
+            ).one()
+            assert row["count"] == 0, f"something in table {table}"
+
 
 @pytest.mark.cassandra
 class TestCassandraStorageGeneratedData(_TestStorageGeneratedData):
@@ -711,6 +703,7 @@ class TestCassandraStorageGeneratedData(_TestStorageGeneratedData):
         pass
 
 
+@pytest.mark.cassandra
 @pytest.mark.parametrize(
     "allow_overwrite,object_type",
     itertools.product(
