@@ -28,11 +28,12 @@ from swh.storage.cassandra.storage import CassandraStorage
 from swh.storage.exc import StorageArgumentException
 
 from .storage_data import StorageData
-from .test_cassandra import (  # noqa, needed for swh_storage fixture
-    cassandra_cluster,
-    keyspace,
-    swh_storage_backend_config,
-)
+
+
+@pytest.fixture
+def swh_storage_backend_config(swh_storage_cassandra_backend_config):
+    return swh_storage_cassandra_backend_config
+
 
 ##############################
 # Common structures
@@ -102,6 +103,7 @@ class CqlRunnerWithXor(CqlRunner):
     )
 
 
+@pytest.mark.cassandra
 def test_add_content_column(
     swh_storage: CassandraStorage, swh_storage_backend_config, mocker  # noqa
 ) -> None:
@@ -235,6 +237,7 @@ class CqlRunnerWithXorPK(CqlRunner):
     )
 
 
+@pytest.mark.cassandra
 def test_change_content_pk(
     swh_storage: CassandraStorage, swh_storage_backend_config, mocker  # noqa
 ) -> None:
