@@ -23,7 +23,9 @@ from swh.storage.tests.storage_tests import TestStorage as _TestStorage
 def swh_storage_backend_config():
     yield {
         "cls": "memory",
-        "journal_writer": {"cls": "memory",},
+        "journal_writer": {
+            "cls": "memory",
+        },
     }
 
 
@@ -110,6 +112,10 @@ class TestInMemoryStorage(_TestStorage):
     def test_origin_count(self):
         pass
 
+    @pytest.mark.skip("in-memory backend has no timeout")
+    def test_querytimeout(self):
+        pass
+
 
 class TestInMemoryStorageGeneratedData(_TestStorageGeneratedData):
     @pytest.mark.skip("Not supported by Cassandra")
